import xml.etree.ElementTree as ET
import argparse
import os

##parses the directory path to allow the file to be renamed
def file_prefixer(name,output,prefix):
    (head,tail)=os.path.split(name)
    tail=prefix+tail
    return os.path.join(output,tail)

##Finds the index of an XML element in a list of elements
##Finds first appearence on default, but can look for later ones
def indexGet(elements, tag, instance=0):
    elem=elements.findall(tag)
    index=elements.getchildren().index(elem[instance])
    return index

class CommentedTreeBuilder(ET.XMLTreeBuilder ):
    def __init__ ( self, html = 0, target = None ):
        ET.XMLTreeBuilder.__init__( self, html, target )
        self._parser.CommentHandler = self.handle_comment
    
    def handle_comment(self, data):
        self._target.start(ET.Comment, {})
        self._target.data(data)
        self._target.end(ET.Comment)

class recipeMaker():
    x=[str(y) for y in range (1,8)]
    w=[str(z) for z in range (9,28)]
    deadVolume=dict.fromkeys(x,15)
    chemNames=["FirstBase", "CompleteCycle", "EndDeblock", "PR2Flush", "SMXFlush", "End"] #List of non-pairedend steps
    for z in w:
        deadVolume[z]=55
    def __init__(self, xml, lanes, addpulse, reagEst, output=None):
        self.xml=xml
        self.recipe=ET.parse(xml, parser=CommentedTreeBuilder())
        self.lanes=lanes
        self.addpulse=addpulse
        self.output=output
        self.reagEst=reagEst

    ##Finds new solution volumes
    def newVolumes(self, IV, x):
        return int(IV)+int(round(self.deadVolume[x]*((8.0/len(self.lanes))-1)))
    ##Finds new solution volume for 1 lane scenario
    def specVolumes(self, IV, x):
        return int(IV)+self.deadVolume[x]*3
    ##Undoes calculation to find actual volume used
    def reverseVolumes(self, CV, x):
        return round(int(CV)*(len(self.lanes)/8.0))

    ##Adjusts solution volumes, checks for a read 2 portion and adds pulse if needed
    def newRecipe(self):
        chem=self.recipe.findall("ChemistryDefinitions/*")
        y=0
        pEnd=False            
        for s in chem:
            x=None
            if s.tag==ET.Comment:
                if 'RESYNTHESIS' in s.text:
                    y=len(self.lanes)
                    pEnd=True
                else:
                    y=0
            elif s.get('Name')!='Resynthesis':
                z=0 #Index of steps in each Chemistry
                steps=s.getiterator()
                for action in steps:
                    if action.tag=="PumpToFlowcell":
                        IV=action.get("Volume")
                        if x!=action.get("Solution"):
                            x=action.get("Solution")
                            if y==1:
                                sv=self.specVolumes(IV,x)
                            else:
                                sv=self.newVolumes(IV, x)
                            if self.addpulse:
                                if s.get('Name') in self.chemNames:
                                    mix=ET.Element("PumpToFlowcell", attrib=action.attrib)
                                    mix.set("Volume", "20")
                                    sv+=40
                                    s.insert(z,ET.Element("Wait", attrib={"Duration":"5000"}))
                                    z+=1
                                    s.insert(z,mix)
                                    z+=1
                            action.set("Volume", str(sv))
                    z+=1
            else:
                ptfc=s.getiterator("PumpToFlowcell")
                for sol in ptfc:
                    if sol.get("Solution")=="13":
                        sol.set("Volume", "54")
                    else:
                        sol.set("Volume", "42")
        return pEnd
    
    ##Finds time needed to run
    def timeEstimater(self, l):
        chemtimes={'Ruu':0}
        chemsteps=self.recipe.findall("ChemistryDefinitions/Chemistry")
        for step in chemsteps:
            time=0
            for s in step.findall("*"):
                if s.tag=="PumpToFlowcell":
                    time+=float(s.get("Volume"))/float(s.get("AspirationRate"))
                else:
                    time+=float(s.get("Duration",default=0))/60000
            if 'Prime' in step.get('Name'):
                chemtimes[step.get('Name')]=1.5
            else:
                chemtimes[step.get('Name')]=time
        totalTime=0
        for p in self.recipe.findall("Protocol/ChemistryRef"):
            totalTime+=chemtimes[p.get("Name",default='Ruu')]
        for p in self.recipe.findall("Protocol/Incorporation"):
            totalTime+=chemtimes[p.get("ChemistryName",default='Ruu')]
        tEst = ' It will take approximately '+ str(int(round(totalTime/60)))+ ' hours to run. '
        ##Only necessary if there is a turnaround
        if l:
            turnaround=0
            cRef=self.recipe.findall('Protocol/ChemistryRef')
            d=cRef.index(self.recipe.find("Protocol/ChemistryRef/[@Name='Denaturation']"))
            r=cRef.index(self.recipe.find("Protocol/ChemistryRef/[@Name='Rehybridisation']"))+1
            for tas in cRef[d:r]:
                turnaround+=chemtimes[tas.get('Name', default='Ruu')]
            turnaround+=15
            tEst=tEst+'The turnaround will take about '\
                  +str(int(turnaround/60)) +' hours and '+str(int(round(turnaround%60,-1)))+' minutes. '
        return tEst

    ##Finds reagent use
    def reagentEstimater(self, elements):
        basevol={'Bre':{}}
        totalvol={'www':0}
        chemsteps=self.recipe.findall("ChemistryDefinitions/Chemistry")
        for step in chemsteps:
            chemname=step.get('Name')
            basevol[chemname]={}
            solutionlist=basevol[chemname]
            for s in step.findall("*"):
                solutionlist[s.get('Solution',default='www')]=solutionlist.setdefault(s.get('Solution',default='www'),0)+float(s.get('Volume',default=0))
        
        for p in elements:
            dicOfsol={}
            if p.tag=="ChemistryRef":
                dicOfsol=basevol[p.get('Name',default='Bre')]
            elif p.tag=="Incorporation":
                dicOfsol=basevol[p.get('ChemistryName',default='Bre')]
            for solution, volume in dicOfsol.items():
                totalvol[solution]=totalvol.setdefault(solution,0)+self.reverseVolumes(volume, solution)
                           
##        for p in self.recipe.findall("Protocol/ChemistryRef"):
##            dicOfsol=basevol[p.get('Name',default='Bre')]
##            for solution, volume in dicOfsol.items():
##                print solution, volume
##                totalvol[solution]=totalvol.setdefault(solution,0)+volume
##        for p in self.recipe.findall("Protocol/Incorporation"):
##            dicOfsol=basevol[p.get('ChemistryName',default='Bre')]
##            for solution, volume in dicOfsol.items():
##                totalvol[solution]=totalvol.setdefault(solution,0)+volume
##                
        del totalvol['www']
        return totalvol
                
                    

##    def extendedPulse(self, messages):
##            
##
##    def singleLane(self, messages):
##        for s in messages:
##            if 'Load pre-weighed Read 2 Resynthesis Reagents' in s.get('Message'):
##                (first, period, last)=s.get('Message').partition('.')
##                s.set("Message",first+'. Please attach extra pump.'+last)
##            if 'Click OK to Start Read 2 first base' in s.get('Message'):
##                (first, period, last)=s.get('Message').partition('.')
##                s.set("Message",first+'. Please disconnect extra pump.'+last)

    ##adds rinse with water if necessary
    def turnaroundPulse(self, r2):
        if self.addpulse and r2:
            ChemistryDef=self.recipe.find("ChemistryDefinitions")
            RinseStep=ET.Element("Chemistry", attrib={"Name":"WaterRinse"})
            ET.SubElement(RinseStep, "PumpToFlowcell", attrib={"Solution":"2", "AspirationRate":"250", \
                                                                 "DispenseRate":"2500", "Volume":"200"})
            ChemistryDef.append(RinseStep)
            protocol=self.recipe.find("Protocol")
            rehybridise=protocol.find("ChemistryRef/[@Name='Rehybridisation']")#finding rehybridization step
            ri=protocol.getchildren().index(rehybridise)
            reconnect=ET.Element("Userwait", attrib={"Message":"Please reconnect all syringes."})
            rinse=ET.Element("ChemistryRef", attrib={"Name":"WaterRinse"})
            disconnect=ET.Element("Userwait", attrib={"Message":"Please disconnect syringes not attached to active lanes."})
            protocol.insert(ri+1,reconnect)
            protocol.insert(ri+2,rinse)
            protocol.insert(ri+3,disconnect)

    def newXML(self):
        r2=self.newRecipe()
        self.turnaroundPulse(r2)
        if self.output!=None:
            root=self.output
        else:
            root=os.path.dirname(self.xml)
        tiles=self.recipe.find("TileSelection").find("Incorporation")
        LI=tiles.getiterator("Lane")
        for qb in LI:
            if qb.get("Index") not in self.lanes[:]:
                tiles.remove(qb)
        x=self.lanes[0]
        for l in range(1,len(self.lanes)):
            x=x+', '+self.lanes[l]
        prefix="Partial_"+str(len(self.lanes))+"lanes_"+x+"_"
        discwarn="disconnect all pumps except for #"
        message=self.recipe.find("Protocol/UserWait")
        (a,b,c)=message.get('Message').partition('.')
        if len(self.lanes)==1:
            synth=self.recipe.findall("Protocol/UserWait")
            for s in synth:
                if 'Load pre-weighed Read 2 Resynthesis Reagents' in s.get('Message'):
                    (first, period, last)=s.get('Message').partition('.')
                    s.set("Message",first+'. Please attach extra pump.'+last)
                if 'Click OK to Start Read 2 first base' in s.get('Message'):
                    (first, period, last)=s.get('Message').partition('.')
                    s.set("Message",first+'. Please disconnect extra pump.'+last)
        if self.addpulse:
            prefix=prefix+"ExtendedVol40_WithPulse_"
            discwarn="Please pre-charge the flowcell with reagent 2 (water) and then "+discwarn
            precharge=self.recipe.find("Protocol/UserWait/[@Message='Load Read 2 SBS Reagents onto the Genome Analyzer. Click OK to Start Read 2 First base.']")
            m=precharge.get('Message')
            m="Pre-charge the flowcell with reagent 2 (water). "+m
            precharge.set("Message", m)
        else:
            discwarn="Please "+discwarn
        message.set("Message",discwarn+x+'. '+a+b+self.timeEstimater(r2)+c)

        if self.reagEst:
            reads=self.readFinder(r2)
            reagentext=open(os.path.join(root, "Reagent_estimate.txt"), 'w')
            read1vol=self.reagentEstimater(reads['read1'])
            reagentext.write('READ 1 VOLUMES'.center(19))
            reagentext.write('\n')
            for solution, volume in read1vol.items():
                reagentext.write('{0:15}{1}\n'.format(solution,volume))
            if r2:
                turnavol=self.reagentEstimater(reads['turnaround'])
                reagentext.write('TURNAROUND VOLUMES'.center(19))
                reagentext.write('\n')
                for solution, volume in turnavol.items():
                    reagentext.write('{0:15}{1}\n'.format(solution,volume))
                read2vol=self.reagentEstimater(reads['read2'])
                reagentext.write('READ 2 VOLUMES'.center(19))
                reagentext.write('\n')
                for solution, volume in read2vol.items():
                    reagentext.write('{0:15}{1}\n'.format(solution,volume))
        self.recipe.write(file_prefixer(self.xml,root,prefix))

    def readFinder(self, r2):
        ##finds first read location
        proslice={}
        protocol=self.recipe.findall("Protocol/*")
        for x in protocol:
            if x.tag==ET.Comment and "READ 1 SBS" in x.text:
                read1index=protocol.index(x)
        if r2:
            for q in protocol:
                if q.tag==ET.Comment and "PREPARATION / RESYNTHESIS" in q.text:
                    turnaroundindex=protocol.index(q)
                elif q.tag==ET.Comment and "READ 2 SBS" in q.text:
                    read2index=protocol.index(q)
            proslice['read1']=protocol[read1index:turnaroundindex]
            proslice['turnaround']=protocol[turnaroundindex:read2index]
            proslice['read2']=protocol[read2index:]
        else:
            proslice['read1']=protocol[read1index:]
        return proslice
            
            
            

parser=argparse.ArgumentParser(description='Adjusts the recipe to account for number of lanes used')
parser.add_argument('xml', help='the recipe file')
parser.add_argument('--pulseon', action='store_true', help = 'adds extended extended pulse for long runs [DEFAULT:FALSE]')
parser.add_argument('--output', help = 'changes the output directory [DEFAULT:INPUT DIRECTORY]')
parser.add_argument('--estimate', action='store_true', help = 'outputs estimated volume usage in Reagent_estimate.txt [DEFAULT:FALSE]')
parser.add_argument('lanes', nargs='+', type=str, choices=['1','2','3','4','5','6','7','8'], help='all lanes being used')
args=parser.parse_args()
recipeMaker(args.xml,args.lanes,args.pulseon, args.estimate,args.output).newXML()
        

